import argparse

from .. import GoodConf


def argparser_add_argument(parser: argparse.ArgumentParser, config: GoodConf):
    """Adds argument for config to existing argparser"""
    help = "Config file."
    cfg = config.model_config
    if cfg.get("file_env_var"):
        help += (
            "Can also be configured via the environment variable: "
            f"{cfg['file_env_var']}"
        )
    if cfg.get("default_files"):
        files_str = ", ".join(cfg["default_files"])
        help += f" Defaults to the first file that exists from [{files_str}]."
    parser.add_argument("-C", "--config", metavar="FILE", help=help)



import json

from goodconf import GoodConf

from .utils import env_var


def test_conf_env_var(mocker, tmpdir):
    mocked_load_config = mocker.patch("goodconf._load_config")
    path = tmpdir.join("myapp.json")
    path.write("")

    class G(GoodConf):
        model_config = {"file_env_var": "CONF"}

    with env_var("CONF", str(path)):
        g = G()
        g.load()
    mocked_load_config.assert_called_once_with(str(path))


def test_conflict(tmpdir):
    path = tmpdir.join("myapp.json")
    path.write(json.dumps({"A": 1, "B": 2}))

    class G(GoodConf):
        A: int
        B: int

        model_config = {"default_files": [path]}

    with env_var("A", "3"):
        g = G()
        g.load()
    assert g.A == 3
    assert g.B == 2


def test_all_env_vars(mocker):
    mocked_set_values = mocker.patch("goodconf.BaseSettings.__init__")
    mocked_load_config = mocker.patch("goodconf._load_config")

    class G(GoodConf):
        pass

    g = G()
    g.load()
    mocked_set_values.assert_called_once_with()
    mocked_load_config.assert_not_called()


def test_provided_file(mocker, tmpdir):
    mocked_load_config = mocker.patch("goodconf._load_config")
    path = tmpdir.join("myapp.json")
    path.write("")

    class G(GoodConf):
        pass

    g = G()
    g.load(str(path))
    mocked_load_config.assert_called_once_with(str(path))


def test_provided_file_from_init(mocker, tmpdir):
    mocked_load_config = mocker.patch("goodconf._load_config")
    path = tmpdir.join("myapp.json")
    path.write("")

    class G(GoodConf):
        pass

    g = G(config_file=str(path))
    g.load()
    mocked_load_config.assert_called_once_with(str(path))


def test_default_files(mocker, tmpdir):
    mocked_load_config = mocker.patch("goodconf._load_config")
    path = tmpdir.join("myapp.json")
    path.write("")
    bad_path = tmpdir.join("does-not-exist.json")

    class G(GoodConf):
        model_config = {"default_files": [str(bad_path), str(path)]}

    g = G()
    g.load()
    mocked_load_config.assert_called_once_with(str(path))



"""
Transparently load variables from environment or JSON/YAML file.
"""

# Note: the following line is included to ensure Python3.9 compatibility.
from __future__ import annotations

import errno
import json
import logging
import os
import sys
from functools import partial
from io import StringIO
from types import GenericAlias
from typing import TYPE_CHECKING, cast, get_args

from pydantic._internal._config import config_keys
from pydantic.fields import Field, PydanticUndefined
from pydantic.main import _object_setattr
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)

if TYPE_CHECKING:
    from typing import Any

    from pydantic.fields import FieldInfo


__all__ = ["GoodConf", "GoodConfConfigDict", "Field"]

log = logging.getLogger(__name__)


class GoodConfConfigDict(SettingsConfigDict):
    # configuration file to load
    file_env_var: str | None
    # if no file is given, try to load a configuration from these files in order
    default_files: list[str] | None


# Note: code from pydantic-settings/pydantic_settings/main.py:
# Extend `config_keys` by pydantic settings config keys to
# support setting config through class kwargs.
# Pydantic uses `config_keys` in `pydantic._internal._config.ConfigWrapper.for_model`
# to extract config keys from model kwargs, So, by adding pydantic settings keys to
# `config_keys`, they will be considered as valid config keys and will be collected
# by Pydantic.
config_keys |= set(GoodConfConfigDict.__annotations__.keys())


def _load_config(path: str) -> dict[str, Any]:
    """
    Given a file path, parse it based on its extension (YAML, TOML or JSON)
    and return the values as a Python dictionary. JSON is the default if an
    extension can't be determined.
    """
    __, ext = os.path.splitext(path)
    if ext in [".yaml", ".yml"]:
        import ruamel.yaml

        yaml = ruamel.yaml.YAML(typ="safe", pure=True)
        loader = yaml.load
    elif ext == ".toml":
        try:
            import tomllib

            def load(stream):
                return tomllib.loads(f.read())
        except ImportError:  # Fallback for Python < 3.11
            import tomlkit

            def load(stream):
                return tomlkit.load(f).unwrap()

        loader = load

    else:
        loader = json.load
    with open(path) as f:
        config = loader(f)
    return config or {}


def _find_file(filename: str, require: bool = True) -> str | None:
    if not os.path.exists(filename):
        if not require:
            return None
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)
    return os.path.abspath(filename)


def _fieldinfo_to_str(field_info: FieldInfo) -> str:
    """
    Return the string representation of a pydantic.fields.FieldInfo.
    """
    if isinstance(field_info.annotation, type) and not isinstance(
        field_info.annotation, GenericAlias
    ):
        # For annotation like <class 'int'>, we use its name ("int").
        field_type = field_info.annotation.__name__
    else:
        if str(field_info.annotation).startswith("typing."):
            # For annotation like typing.Literal['a', 'b'], we use
            # its string representation, but without "typing." ("Literal['a', 'b']").
            field_type = str(field_info.annotation)[len("typing.") :]
        else:
            # For annotation like list[str], we use its string
            # representation ("list[str]").
            field_type = field_info.annotation
    return field_type


def initial_for_field(name: str, field_info: FieldInfo) -> Any:
    try:
        json_schema_extra = field_info.json_schema_extra or {}
        if not callable(json_schema_extra["initial"]):
            raise ValueError(f"Initial value for `{name}` must be a callable.")
        return field_info.json_schema_extra["initial"]()
    except KeyError:
        if (
            field_info.default is not PydanticUndefined
            and field_info.default is not ...
        ):
            return field_info.default
        if field_info.default_factory is not None:
            return field_info.default_factory()
    if type(None) in get_args(field_info.annotation):
        return None
    return ""


class FileConfigSettingsSource(PydanticBaseSettingsSource):
    """
    Source class for loading values provided during settings class initialization.
    """

    def __init__(self, settings_cls: type[BaseSettings]):
        super().__init__(settings_cls)

    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> tuple[Any, str, bool]:
        # Nothing to do here. Only implement the return statement to make mypy happy
        return None, "", False

    def __call__(self) -> dict[str, Any]:
        settings = cast(GoodConf, self.settings_cls)
        selected_config_file = None
        if cfg_file := self.current_state.get("_config_file"):
            selected_config_file = cfg_file
        elif (file_env_var := settings.model_config.get("file_env_var")) and (
            cfg_file := os.environ.get(file_env_var)
        ):
            selected_config_file = _find_file(cfg_file)
        else:
            for filename in settings.model_config.get("default_files") or []:
                selected_config_file = _find_file(filename, require=False)
                if selected_config_file:
                    break
        if selected_config_file:
            values = _load_config(selected_config_file)
            log.info("Loading config from %s", selected_config_file)
        else:
            values = {}
            log.info("No config file specified. Loading with environment variables.")
        return values

    def __repr__(self) -> str:
        return "FileConfigSettingsSource()"


class GoodConf(BaseSettings):
    def __init__(
        self, load: bool = False, config_file: str | None = None, **kwargs
    ) -> None:
        """
        :param load: load config file on instantiation [default: False].

        A docstring defined on the class should be a plain-text description
        used as a header when generating a configuration file.
        """
        if kwargs or load:  # Emulate Pydantic behavior, load immediately
            self._load(_init_config_file=config_file, **kwargs)
        elif config_file:
            _object_setattr(
                self, "_load", partial(self._load, _init_config_file=config_file)
            )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """Load environment variables before init"""
        return (
            init_settings,
            env_settings,
            dotenv_settings,
            FileConfigSettingsSource(settings_cls),
            file_secret_settings,
        )

    model_config = GoodConfConfigDict()

    def _settings_build_values(
        self,
        init_kwargs: dict[str, Any],
        **kwargs,
    ) -> dict[str, Any]:
        state = super()._settings_build_values(
            init_kwargs,
            **kwargs,
        )
        state.pop("_config_file", None)
        return state

    def _load(
        self,
        _config_file: str | None = None,
        _init_config_file: str | None = None,
        **kwargs,
    ):
        if config_file := _config_file or _init_config_file:
            kwargs["_config_file"] = config_file
        super().__init__(**kwargs)

    def load(self, filename: str | None = None) -> None:
        self._load(_config_file=filename)

    @classmethod
    def get_initial(cls, **override) -> dict:
        return {
            k: override.get(k, initial_for_field(k, v))
            for k, v in cls.model_fields.items()
        }

    @classmethod
    def generate_yaml(cls, **override) -> str:
        """
        Dumps initial config in YAML
        """
        import ruamel.yaml

        yaml = ruamel.yaml.YAML()
        yaml.representer.add_representer(
            type(None),
            lambda self, d: self.represent_scalar("tag:yaml.org,2002:null", "~"),
        )
        yaml_str = StringIO()
        yaml.dump(cls.get_initial(**override), stream=yaml_str)
        yaml_str.seek(0)
        dict_from_yaml = yaml.load(yaml_str)
        if cls.__doc__:
            dict_from_yaml.yaml_set_start_comment("\n" + cls.__doc__ + "\n\n")
        for k in dict_from_yaml.keys():
            if cls.model_fields[k].description:
                description = cast(str, cls.model_fields[k].description)
                dict_from_yaml.yaml_set_comment_before_after_key(
                    k, before="\n" + description
                )
        yaml_str = StringIO()
        yaml.dump(dict_from_yaml, yaml_str)
        yaml_str.seek(0)
        return yaml_str.read()

    @classmethod
    def generate_json(cls, **override) -> str:
        """
        Dumps initial config in JSON
        """
        return json.dumps(cls.get_initial(**override), indent=2)

    @classmethod
    def generate_toml(cls, **override) -> str:
        """
        Dumps initial config in TOML
        """
        import tomlkit
        from tomlkit.items import Item

        toml_str = tomlkit.dumps(cls.get_initial(**override))
        dict_from_toml = tomlkit.loads(toml_str)
        document = tomlkit.document()
        if cls.__doc__:
            document.add(tomlkit.comment(cls.__doc__))
        for k, v in dict_from_toml.unwrap().items():
            document.add(k, v)
            if cls.model_fields[k].description:
                description = cast(str, cls.model_fields[k].description)
                cast(Item, document[k]).comment(description)
        return tomlkit.dumps(document)

    @classmethod
    def generate_markdown(cls) -> str:
        """
        Documents values in markdown
        """
        lines = []
        if cls.__doc__:
            lines.extend([f"# {cls.__doc__}", ""])

        for k, field_info in cls.model_fields.items():
            lines.append(f"* **{k}**")
            if field_info.is_required():
                lines[-1] = f"{lines[-1]} _REQUIRED_"
            if field_info.description:
                lines.append(f"  * description: {field_info.description}")
            # We want to append a line with the field_info type, and sometimes
            # field_info.annotation looks the way we want, like 'list[str]', but
            # other times, it includes some extra text, like '<class 'bool'>'.
            # Therefore, we have some logic to make the type show up the way we want.
            field_type = _fieldinfo_to_str(field_info)
            lines.append(f"  * type: `{field_type}`")
            if field_info.default not in [None, PydanticUndefined]:
                lines.append(f"  * default: `{field_info.default}`")
        return "\n".join(lines)

    def django_manage(self, args: list[str] | None = None):
        args = args or sys.argv
        from .contrib.django import execute_from_command_line_with_config

        execute_from_command_line_with_config(self, args)



import json
import os
import re
from textwrap import dedent
from typing import Optional, List, Literal

import pytest
from pydantic import Field, ValidationError
from pydantic.fields import FieldInfo

from goodconf import GoodConf, FileConfigSettingsSource
from tests.utils import env_var


def test_initial():
    class TestConf(GoodConf):
        a: bool = Field(initial=lambda: True)
        b: bool = Field(default=False)

    initial = TestConf.get_initial()
    assert len(initial) == 2
    assert initial["a"] is True
    assert initial["b"] is False


def test_dump_json():
    class TestConf(GoodConf):
        a: bool = Field(initial=lambda: True)

    assert TestConf.generate_json() == '{\n  "a": true\n}'
    assert TestConf.generate_json(not_a_value=True) == '{\n  "a": true\n}'
    assert TestConf.generate_json(a=False) == '{\n  "a": false\n}'


def test_dump_toml():
    pytest.importorskip("tomlkit")

    class TestConf(GoodConf):
        a: bool = False
        b: str = "Happy"

    output = TestConf.generate_toml()
    assert "a = false" in output
    assert 'b = "Happy"' in output

    class TestConf(GoodConf):
        "Configuration for My App"

        a: str = Field(description="this is a")
        b: str

    output = TestConf.generate_toml()

    assert "# Configuration for My App\n" in output
    assert 'a = "" # this is a' in output
    assert 'b = ""' in output


def test_dump_yaml():
    pytest.importorskip("ruamel.yaml")

    class TestConf(GoodConf):
        "Configuration for My App"

        a: str = Field(description="this is a")
        b: str

    output = TestConf.generate_yaml()
    output = re.sub(r" +\n", "\n", output)
    assert "\n# Configuration for My App\n" in output
    assert (
        dedent(
            """\
        # this is a
        a: ''
        """
        )
        in output
    )
    assert "b: ''" in output

    output_override = TestConf.generate_yaml(b="yes")
    assert "a: ''" in output_override
    assert "b: yes" in output_override


def test_dump_yaml_no_docstring():
    pytest.importorskip("ruamel.yaml")

    class TestConf(GoodConf):
        a: str = Field(description="this is a")

    output = TestConf.generate_yaml()
    output = re.sub(r" +\n", "\n", output)
    assert output == dedent(
        """
        # this is a
        a: ''
        """
    )


def test_dump_yaml_none():
    pytest.importorskip("ruamel.yaml")

    class TestConf(GoodConf):
        a: Optional[str]

    output = TestConf.generate_yaml()
    assert output.strip() == "a: ~"


def test_generate_markdown():
    help_ = "this is a"

    class TestConf(GoodConf):
        "Configuration for My App"

        a: int = Field(description=help_, default=None)
        b: int = Field(description=help_, default=5)
        c: str

    mkdn = TestConf.generate_markdown()
    # Not sure on final format, just do some basic smoke tests
    assert TestConf.__doc__ in mkdn
    assert help_ in mkdn


def test_generate_markdown_no_docstring():
    help_ = "this is a"

    class TestConf(GoodConf):
        a: int = Field(description=help_, default=5)
        b: str

    mkdn = TestConf.generate_markdown()
    # Not sure on final format, just do some basic smoke tests
    assert f"  * description: {help_}" in mkdn.splitlines()


def test_generate_markdown_default_false():
    class TestConf(GoodConf):
        a: bool = Field(default=False)

    lines = TestConf.generate_markdown().splitlines()
    assert "  * type: `bool`" in lines
    assert "  * default: `False`" in lines


def test_generate_markdown_types():
    class TestConf(GoodConf):
        a: Literal["a", "b"] = Field(default="a")
        b: list[str] = Field()
        c: None

    lines = TestConf.generate_markdown().splitlines()
    assert "  * type: `Literal['a', 'b']`" in lines
    assert "  * type: `list[str]`" in lines
    assert "default: `PydanticUndefined`" not in str(lines)


def test_generate_markdown_required():
    class TestConf(GoodConf):
        a: str

    lines = TestConf.generate_markdown().splitlines()
    assert "* **a** _REQUIRED_" in lines


def test_undefined():
    c = GoodConf()
    with pytest.raises(AttributeError):
        c.UNDEFINED


def test_required_missing():
    class TestConf(GoodConf):
        a: str = Field()

    c = TestConf()

    with pytest.raises(ValidationError):
        c.load()

    with pytest.raises(ValidationError):
        TestConf(load=True)


def test_set_on_init():
    class TestConf(GoodConf):
        a: str = Field()

    val = "test"
    c = TestConf(a=val)
    assert c.a == val


def test_env_prefix():
    class TestConf(GoodConf):
        a: bool = False

        model_config = {"env_prefix": "PREFIX_"}

    with env_var("PREFIX_A", "True"):
        c = TestConf(load=True)

    assert c.a


def test_precedence(tmpdir):
    path = tmpdir.join("myapp.json")
    path.write(json.dumps({"init": "file", "env": "file", "file": "file"}))

    class TestConf(GoodConf, default_files=[path]):
        init: str = ""
        env: str = ""
        file: str = ""

    os.environ["INIT"] = "env"
    os.environ["ENV"] = "env"
    try:
        c = TestConf(init="init")
        assert c.init == "init"
        assert c.env == "env"
        assert c.file == "file"
    finally:
        del os.environ["INIT"]
        del os.environ["ENV"]


def test_fileconfigsettingssource_repr():
    class SettingsClass:
        model_config = {}

    fileconfigsettingssource = FileConfigSettingsSource(SettingsClass)

    assert repr(fileconfigsettingssource) == "FileConfigSettingsSource()"

    field = FieldInfo(title="testfield")


def test_fileconfigsettingssource_get_field_value():
    class SettingsClass:
        model_config = {}

    fileconfigsettingssource = FileConfigSettingsSource(SettingsClass)
    field = FieldInfo(title="testfield")
    assert fileconfigsettingssource.get_field_value(field, "testfield") == (
        None,
        "",
        False,
    )
    assert fileconfigsettingssource.get_field_value(None, "a") == (None, "", False)



from typing import Optional

import pytest

from goodconf import Field, GoodConf, initial_for_field

from .utils import KEY


def test_initial():
    class C(GoodConf):
        f: str = Field(initial=lambda: "x")

    assert initial_for_field(KEY, C.model_fields["f"]) == "x"


def test_initial_bad():
    class C(GoodConf):
        f: str = Field(initial="x")

    with pytest.raises(ValueError):
        initial_for_field(KEY, C.model_fields["f"])


def test_initial_default():
    class C(GoodConf):
        f: str = Field("x")

    assert initial_for_field(KEY, C.model_fields["f"]) == "x"


def test_initial_default_factory():
    class C(GoodConf):
        f: str = Field(default_factory=lambda: "y")

    assert initial_for_field(KEY, C.model_fields["f"]) == "y"


def test_no_initial():
    class C(GoodConf):
        f: str = Field()

    assert initial_for_field(KEY, C.model_fields["f"]) == ""


def test_default_initial():
    """Can get initial when Field is not used"""

    class G(GoodConf):
        a: str = "test"

    initial = G().get_initial()
    assert initial["a"] == "test"


def test_optional_initial():
    class G(GoodConf):
        a: Optional[str]

    initial = G().get_initial()
    assert initial["a"] is None
